// Copyright 2001  EPITA Research and Development Laboratory
//
// This file is part of the Olena Library.  This library is free
// software; you can redistribute it and/or modify it under the terms
// of the GNU General Public License version 2 as published by the
// Free Software Foundation.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this library; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// As a special exception, you may use this file as part of a free
// software library without restriction.  Specifically, if other files
// instantiate templates or use macros or inline functions from this
// file, or you compile this file and link it with other files to
// produce an executable, this file does not by itself cause the
// resulting executable to be covered by the GNU General Public
// License.  This exception does not however invalidate any other
// reasons why the executable file might be covered by the GNU General
// Public License.

#ifndef OLENA_MORPHO_FAST_MORPHO_HXX
# define OLENA_MORPHO_FAST_MORPHO_HXX


namespace oln {
  namespace morpho {

    // Update HIST by adding elements from _SE_ADD, and removing
    // those from _SE_REM.
    template<class _I, class _E1, class _E2>
    inline void
    hist_update(utils::histogram<Value(_I)>& hist,
		const image<_I>& _input,
		const Point(_I)& p,
		const struct_elt<_E1>& _se_rem,
		const struct_elt<_E2>& _se_add)
    {
      Exact_cref(I, input);
      {
	Exact_cref(E1, se_rem);
	Iter(E1) dp(se_rem);
	for_all(dp)
	  --hist[input[p + dp]];
      }
      {
	Exact_cref(E2, se_add);
	Iter(E2) dp(se_add);
	for_all(dp)
	  ++hist[input[p + dp]];
      }
    }

    // Find structuring elements to be added/removed from the histogram
    // when we move forward along each direction.
    template<class _E1, class _E2, class _E3>
    void find_struct_elts(const struct_elt<_E1>& _se,
			  struct_elt<_E2> _se_add[type::exact<_E1>::ret::dim],
			  struct_elt<_E3> _se_rem[type::exact<_E1>::ret::dim])
    {
      Exact_cref(E1, se);
      Exact_ptr(E2, se_add);
      Exact_ptr(E3, se_rem);
      const unsigned dim = E1::dim;

      // back[n] allows to move backward on coordinate `n'.
      DPoint(E1) back[dim];
      for (unsigned n = 0; n < dim; ++n)
	back[n].nth(n) = -1;

      Iter(E1) dp(se);
      Iter(E1) dp_prime(se);

      for_all(dp)
	{
	  bool add[dim];	// whether to add `dp' when moving forward
				// on coordinate `n'.
	  bool rem[dim];	// whether to remove `dp'.
	  for (unsigned n = 0; n < dim; ++n)
	    add[n] = rem[n] = true;

	  for_all(dp_prime)
	    for (unsigned n = 0; n < dim; ++n)
	      {
		if (dp_prime.cur() + back[n] == dp)
		  // DP_PRIME is already in SE: don't add it.
		  add[n] = false;

		if (dp.cur() + back[n] == dp_prime)
		  // DP is still in SE: don't remove it.
		  rem[n] = false;
	      }

	  for (unsigned n = 0; n < dim; ++n)
	    {
	      if (add[n])
		se_add[n].add(dp);
	      if (rem[n])
		se_rem[n].add(dp.cur() + back[n]);
	    }
	}

      for (unsigned n = 0; n < dim; ++n)
	postcondition(se_add[n].card() == se_add[n].card());
    }


    template<class _I, class _E,
	     Value(_I) (*func)(const utils::histogram<Value(_I)>&)>
    Concrete(_I)
    fast_morpho(const image<_I>& _input,
		const struct_elt<_E>& _se)
    {
      Exact_cref(I, input);
      Exact_cref(E, se);
      const unsigned dim = E::dim;

      // prepare output
      Concrete(I) output(input.size());
      border::adapt_copy(input, se.delta());

      // compute delta structuring elements for forward movements
      E se_add[dim];
      E se_rem[dim];
      find_struct_elts(se, se_add, se_rem);

      // compute delta structuring elements for backward movements
      E se_add_back[dim];
      E se_rem_back[dim];
      for (unsigned n = 0; n < dim; ++n)
	for (unsigned i = 0; i < se_add[n].card(); ++i)
	{
	  DPoint(I) dp = se_add[n].dp(i);
	  dp.nth(n) += 1;
	  se_rem_back[n].add(dp);

	  dp = se_rem[n].dp(i);
	  dp.nth(n) += 1;
	  se_add_back[n].add(dp);
	}

      // We will zigzag over the image so that only one coordinate
      // changes at each step.  The path looks as follow on
      // 2D images:
      //   -----------------.
      //   ,----------------'
      //   `----------------.
      //   ,----------------'
      //   `-----------------
      // (The algorithm below handles the n-dimensional case.)

      // For each dimension, backward[n] records whether
      // we are moving backward or not.
      bool backward[dim];
      // Initially, go forward in each direction.
      for (unsigned n = 0; n < dim; ++n)
	backward[n] = false;

      // Initialize the histogram with the values around the first point.
      utils::histogram<Value(I)> hist;
      Point(I) p;

      Iter(E) dp(se);
      for_all(dp)
	++hist[input[p + dp]];

      // Process the image.
      // ------------------

      output[p] = func(hist);	// First point.


      const typename I::image_size size = input.size();
      for(;;)
	{
	  // Move to next position.
	  // FIXME: For now we'll just browse the image
	  // in a slice/row/col order.  It would be faster to select a
	  // browsing order where the se_add[n].card() are decreasing.
	  unsigned n = dim;
	  for (;;)
	    {
	      --n;
	      if (backward[n])
		{
		  --p.nth(n);
		  if (p.nth(n) >= 0)
		    break;
		  p.nth(n) = 0;
		  backward[n] = false;
		}
	      else
		{
		  ++p.nth(n);
		  if (p.nth(n) < size.nth(n))
		    break;
		  p.nth(n) = size.nth(n) - 1;
		  backward[n] = true;
		}

	      if (n == 0) {
		/* If we `return output' here, the final `return' is
		   never reached.  However g++ complains if we remove
		   it: so we need to keep the two returns.

		   Each `return' translate into a fairly large amount
		   of code (destroying each object used).  So we used
		   a `goto' to prevent this duplication of code.

		   More over keeping the `return' code out of the
		   loop reduce it's size, which is good.
		*/
		goto exit_both_loops;
	      }
	    }

	  // Update historgram.
	  if (backward[n])
	    hist_update(hist, input, p, se_rem_back[n], se_add_back[n]);
	  else
	    hist_update(hist, input, p, se_rem[n], se_add[n]);

	  // Compute result value.
	  output[p] = func(hist);
	}

    exit_both_loops:
      return output;
    }

  } // morpho
} // oln

#endif // OLENA_MORPHO_FAST_MORPHO_HXX
