// Copyright 2001  EPITA Research and Development Laboratory
//
// This file is part of the Olena Library.  This library is free
// software; you can redistribute it and/or modify it under the terms
// of the GNU General Public License version 2 as published by the
// Free Software Foundation.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this library; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.
//
// As a special exception, you may use this file as part of a free
// software library without restriction.  Specifically, if other files
// instantiate templates or use macros or inline functions from this
// file, or you compile this file and link it with other files to
// produce an executable, this file does not by itself cause the
// resulting executable to be covered by the GNU General Public
// License.  This exception does not however invalidate any other
// reasons why the executable file might be covered by the GNU General
// Public License.

#ifndef OLENA_UTILS_HISTOGRAM_HH
# define OLENA_UTILS_HISTOGRAM_HH

# include <vector>
# include "basics.hh"

namespace oln {

  namespace utils {

    template< typename T, typename U>
    class histogram;

    template< typename T, typename U >
    T min(const histogram<T, U>& hist);

    template< typename T, typename U >
    T max(const histogram<T, U>& hist);


    template< typename T, typename U = unsigned>
    class histogram
    {
    public:

      histogram() :
	_values(unsigned(T::max().value()
			 - T::min().value()
			 + T::unit().value()))
      {
	for (unsigned i = 0; i < _values.size(); ++i)
	  _values[i] = 0;
      }

      U& operator[](const T& i)
      {
	return _values[unsigned(i.value() - T::min().value())];
      }

      friend T min<T, U>(const histogram<T, U>& hist);
      friend T max<T, U>(const histogram<T, U>& hist);

      template <class _I>
      void init(const image<_I> &_img)
      {
	Exact_cref(I, img);
	Iter(I) p(img);

	for_all (p)
	  _values[unsigned(img[p].value())]++;
      }

    private:
      std::vector< U > _values;
    };


    template< typename T, typename U >
    inline
    T min(const histogram<T, U>& hist)
    {
      unsigned i;
      for (i = 0; i < unsigned(T::max().value() - T::min().value()); ++i)
	if (hist._values[i] > 0)
	  break;
      return T(T::min().value() + i);
    }


    template< typename T, typename U >
    inline
    T max(const histogram<T, U>& hist)
    {
      unsigned i;
      for (i = unsigned(T::max().value() - T::min().value()); i > 0; --i)
	if (hist._values[i] > 0)
	  break;
      return T(T::min().value() + i);
    }


    template<class _I>
    void distrib_sort(const image<_I>& _im, std::vector<Point(_I)> &v)
    {
      Exact_cref(I, im);
      typedef Value(I) val;

      // check the size
      precondition(v.size() == unsigned(im.nrows() * im.ncols()));

      // calculate the histogram of the image
      utils::histogram< Value(I) > histo;
      histo.init(im);

      // initialize the array of pointer to the point in the result
      // with the histogram we can know the number of each color and
      // then calculate an array of pointer for quick access to each
      // value of the image
      std::vector< Point(I)* > ptr(val::max().value() + 1);
      ptr[0] = &(v[0]);
      for (int i = val::min().value() + 1; i <= val::max().value(); i++)
	ptr[unsigned(i)] = ptr[unsigned(i - 1)] + histo[i - 1];

      // Now iterate on the image to sort point in the order of their
      // level
      Iter(I) p(im);
      for_all(p)
	*(ptr[unsigned(im[p].value())]++) = p;
    }

  } // utils

} // oln

#endif // OLENA_UTILS_HISTOGRAM_HH
